#### R commands for "Chequerboard patterns" in 
#### Jie Yang, Klaus Miescke, and Peter McCullagh (2012).
#### Classification Based on a Permanental Process with Cyclic Approximation, 
#### Biometrika, Vol. 99, No. 4, 775-786.
#### http://www.math.uic.edu/~jyang06/publications/per.zip


### version: 03/11/2014
### please send your comments to: Jie Yang at jyang06@uic.edu

### need data set "xc2.dat", "xc2test.dat" in "per.zip"

#### 3*3 chequerboard pattern (two classes) data
#     0 1 0              4 9 5
#     1 0 1 (classes)    7 3 8 (index of sub-squares)
#     0 1 0              1 6 2

#### use permanental classification model

## function to plot the output of "random33"
## input: X[1:n,1:2], xclass[1:n] (taking values in {1,2,3})
## output: two dimensional graph with at most three classes
plotrandom33<-function(X, xclass) {
  plot(c(0,4),c(0,3),xlab="x",ylab="y",type="n");
  tline<-seq(0,3,length=11);
  points(rep(0,11),tline,lty=1,type="l");
  points(rep(1,11),tline,lty=2,type="l");
  points(rep(2,11),tline,lty=2,type="l");
  points(rep(3,11),tline,lty=1,type="l");
  points(tline,rep(0,11),lty=1,type="l");
  points(tline,rep(1,11),lty=2,type="l");
  points(tline,rep(2,11),lty=2,type="l");
  points(tline,rep(3,11),lty=1,type="l");
  points(X[xclass==1,1],X[xclass==1,2],pch=1,type="p");
  points(X[xclass==2,1],X[xclass==2,2],pch=4,type="p");
  points(X[xclass==3,1],X[xclass==3,2],pch=15,type="p");
  if(max(xclass)==2) legend(3.2,3,legend=c("Class 1","Class 2"),pch=c(1,4));
  if(max(xclass)==3) legend(3.2,3,legend=c("Class 1","Class 2","Class 3"),pch=c(1,4,15));
}

## read "xc2.dat"
temp<-read.table("xc2.dat")
X <- rep(0, 90*2)                    # training data
dim(X) <- c(90, 2)
xclass <- rep(0, 90)                 # class labels of training data
for(i in 1:90) {X[i,1]<-temp[i,1];X[i,2]<-temp[i,2];xclass[i]<-temp[i,3];}


## plot "xc2.dat"
par(mfrow=c(1,1))
plotrandom33(X, xclass)


## read "xc2test.dat"
temp<-read.table("xc2test.dat")
tNum <- 60
Xt <- rep(0, tNum*tNum*2)            # testing data
dim(Xt) <- c(tNum*tNum, 2)
tclass <- rep(0, tNum*tNum)          # class labels of training data
for(i in 1:(tNum*tNum)) {Xt[i,1]<-temp[i,1];Xt[i,2]<-temp[i,2];tclass[i]<-temp[i,3];}


## plot "xc2test.dat"
par(mfrow=c(1,1))
plotrandom33(Xt, tclass)


### candidate covariance function
# exponential covariance
K1 <- function(x, tau){
  exp(-abs(x)/tau);
}

# Gaussian covariance
K2 <- function(x, tau){
  exp(-x^2/tau^2);
}

### R function to compute approximated permanent ratios based on R1,R2,R3
### Input: Kmat=K(c(tvec,x),c(tvec,x)) -- Matix with dim (npts+nobs)*(npts+nobs);
###        npts -- number of testing points;  nobs -- number of training points;
###        alpha -- parameter in per_alpha[K](X), default value is 1 which leading to usual permanent
### Output: R1, R2, R3 -- vectors per_alpha[K](X,t)/per_alpha[K](X) based on R1,R2,R3, dim: npts
per.ratio<-function(Kmat, npts, nobs, alpha) {
  n <- nobs + npts;
  if(missing(alpha)) alpha <- 1;
  wt <- diag(Kmat);
  Kmat <- diag(1/sqrt(wt)) %*% Kmat %*% diag(1/sqrt(wt));
  Kxx <- Kmat[(npts+1):n, (npts+1):n];
  Ktx <- Kmat[1:npts, (npts+1):n];
  if(npts==1) Ktx <- t(as.matrix(Ktx));
  kxx2m <- Kxx^2;
  
  R1 <- apply(Ktx^2, 1, sum) + alpha;           ## R1(t;X)
  R1 <- R1 * wt[1:npts];
  
  temp <- kxx2m; diag(temp) <- alpha;
  R1i <- apply(temp, 1, sum);                   ## R1(xi;X-i)
  
  temp <- Kxx;  diag(temp) <- alpha;
  R2 <- ((Ktx %*% temp) * Ktx) %*% (1/R1i);     ## R2(t;X)
  R2 <- as.vector(R2) + alpha;
  R2 <- R2 * wt[1:npts];
  
  R1ij <- matrix(0, nobs, nobs);                ## R1(xj;X-i-j)
  for(i in 1:nobs) {
    temp <- kxx2m; temp[,i] <- 0; diag(temp)<-alpha;
    R1ij[i,] <- apply(temp, 1, sum);
  }
  diag(R1ij) <- R1i;
  
  R2i <- rep(0, nobs);                          ## R2(xi;X-i)
  for(i in 1:nobs) {
    temp <- Kxx; temp[,i] <- 0; diag(temp)<-alpha;
    temp <- (temp %*% Kxx[i,]) * Kxx[i,] / R1ij[i,];
    temp[i] <- alpha;
    R2i[i] <- sum(temp);
  }
  
  temp <- matrix(0, npts, nobs);
  for(i in 1:nobs) {
    temp1 <- Kxx; temp1[i,] <- 0; diag(temp1)<-alpha;
    temp1 <- Ktx %*% temp1;
    temp2 <- Kxx[i,] / R1ij[i,]; temp2[i]<- 0;
    temp[,i] <- temp1 %*% temp2;
  }
  R3 <- ((temp + alpha * Ktx) * Ktx) %*% (1/R2i);    ## R3(t;X)
  R3 <- as.vector(R3) + alpha;
  R3 <- R3 * wt[1:npts];
  list(R1=R1, R2=R2, R3=R3);
}

### R function to compute the predictive distribution for the classes
### Other functions needed: per.ratio
### Input: Kmat, npts, nobs, alpha (see per.ratio),
###        xclass -- class labels, from 1 to k
### Output: p1,...,p_k; RatioSum; Class: 1st largest; Class2: 2nd largest -- based on R1, R2, R3
predict.self <- function(Kmat, npts, nobs, class, alpha) {
  n <- npts + nobs;
  k <- max(class);
  if(missing(alpha)) alpha <- 1;
  ratio <- rep(0,npts*k*3);
  dim(ratio)<-c(npts,k,3);
  for(r in 1:k) {
    nr <- sum(class==r);
    index <- as.logical(c(rep(1, npts), as.numeric(class==r)));
    temp <- per.ratio(Kmat[index, index], npts, nr, alpha);
    ratio[,r,1] <- temp$R1;
    ratio[,r,2] <- temp$R2;
    ratio[,r,3] <- temp$R3;
  }
  ans <- rep(0,npts*(k+3)*3);
  dim(ans) <- c(npts,k+3,3);
  Class <- Class2 <- rep(1,npts);
  for(r in 1:3) {
    RatioSum <- ratio[,,r] %*% rep(1, k);
    for(i in 1:npts) {
      Class[i] <- sort.list(ratio[i,,r])[k];
      Class2[i] <- sort.list(ratio[i,,r])[k-1];
    }
    ans[,,r] <- cbind(ratio[,,r]/(RatioSum %*% t(rep(1, k))), RatioSum, Class, Class2);
  }
  R1 <- ans[,,1]; if(npts==1) R1 <- t(R1); colnames(R1)<-c(seq(1:k),"RatioSum","Class","Class2");
  R2 <- ans[,,2]; if(npts==1) R2 <- t(R2); colnames(R2)<-c(seq(1:k),"RatioSum","Class","Class2");
  R3 <- ans[,,3]; if(npts==1) R3 <- t(R3); colnames(R3)<-c(seq(1:k),"RatioSum","Class","Class2");
  list(R1=R1,R2=R2,R3=R3);
}


### R function to find the estimates for the parameters: alpha, tau for K1, K2
### based on K-fold cross validation: min -2*sum_i log \hat p_{i,l_i}
### Functions needed: per.ratio, predict.self, K1, K2
### Input: X (nobs*nx), class (nobs*1), vector of alpha, vector of tau, K (K-fold), Knum(1--K1, 2--K2, 3--K3, default=2)
### Output: deviance matrix (alpha, tau) based R1,R2,R3
###         error rate matrix (alpha, tau) based R1,R2,R3
Kest1k <- function(X, class, alphavec, tauvec, K, Knum) {
  n <- dim(X)[1];
  nc <- max(class);
  ind <- ceiling(sample(n)/(n/K));
  nk <- rep(0, K);
  for(i in 1:K) nk[i] <- sum(ind==i);
  d <- rep(0, n*n);
  dim(d) <- c(n, n);
  diag(d) <- 0;
  for(i in 1:(n-1)) for(j in (i+1):n) {
    d[i,j]<-d[j,i]<-sqrt(sum((X[i,]-X[j,])^2));
  }
  d <- d/median(d);               # important adjustment
  anum <- length(alphavec);
  taunum <- length(tauvec);
  ans <- rep(0,anum*taunum*3);    # device matrix
  dim(ans) <- c(anum,taunum,3);
  ans1 <- ans;                    # error rate matrix
  class0 <- class;
  for(itau in 1:taunum) {
    #    cat("\n itau=",itau);
    tau <- tauvec[itau];
    if(Knum==1) Kmat0 <- K1(d, tau) else {
      if(Knum==3) Kmat0 <- K3(d, tau) else Kmat0 <- K2(d, tau);
    };
    for(s in 1:K) {
      npts <- nk[s];
      nobs <- n - nk[s];
      xind <- c(seq(1:n)[ind==s],seq(1:n)[ind!=s]);
      Kmat <- Kmat0[xind,xind];
      class <- class0[ind!=s];
      class1 <- class0[ind==s];
      class2 <- cbind(seq(1:npts), class1);
      for(ialpha in 1:anum) {
        alpha <- alphavec[ialpha];
        pre<-predict.self(Kmat,npts,nobs,class,alpha);
        ans[ialpha,itau,1] <- ans[ialpha,itau,1] - 2*sum(log(pre$R1[class2]));
        ans[ialpha,itau,2] <- ans[ialpha,itau,2] - 2*sum(log(pre$R2[class2]));
        ans[ialpha,itau,3] <- ans[ialpha,itau,3] - 2*sum(log(pre$R3[class2]));
        ans1[ialpha,itau,1] <- ans1[ialpha,itau,1] + sum(pre$R1[,nc+2]!=class1);
        ans1[ialpha,itau,2] <- ans1[ialpha,itau,2] + sum(pre$R2[,nc+2]!=class1);
        ans1[ialpha,itau,3] <- ans1[ialpha,itau,3] + sum(pre$R3[,nc+2]!=class1);
      }
    }
  }
  ans1 <- ans1/n;
  list(devicemat=ans,errorrate=ans1);
}


#### use permanent classification method
# use cross-validation to choose parameters "alpha", "tau" based on (X, xclass)
# use function "Kest1k(X, class, alphavec, tauvec, K, Knum)"
K <- 10      # K-fold cross-validation
alphavec <- seq(0, 1.5, by=0.5)
anum <- length(alphavec)

# for K1
tauvec <- seq(0.05, 0.4, by=0.01)
taunum <- length(tauvec)
set.seed(123)
tempt <- proc.time()
ktemp <- Kest1k(X=X, class=xclass, alphavec=alphavec, tauvec=tauvec, K=K, Knum=1)
proc.time()-tempt  #  40.18
ktemp$errorrate[,,3]
ktemp$device[,,3]
contour(alphavec,tauvec,ktemp$devicemat[,,3],xlab="alpha",ylab="tau",
        main="K1 Covariance",nlevels=80)
ktemp.k1 <- ktemp          # save for K1 
temp <- rep(0, anum*4)     # alpha, optimal tau, deviance, error rate
dim(temp) <- c(anum, 4)
temp[,1] <- alphavec
for(i in 1:anum) {
  tauindex <- order(ktemp$devicemat[i,,3])[1];
  temp[i,2] <- tauvec[tauindex];
  temp[i,3] <- ktemp$devicemat[i, tauindex, 3];
  temp[i,4] <- ktemp$errorrate[i, tauindex, 3];
}
colnames(temp) <- c("alpha", "tau", "deviance", "error rate")
taualpha.k1 <- temp
temp
#      alpha  tau deviance error rate
# [1,]   0.0 0.10 42.86389 0.08888889
# [2,]   0.5 0.20 74.82697 0.12222222
# [3,]   1.0 0.23 86.97744 0.12222222
# [4,]   1.5 0.25 94.10602 0.12222222

# for K2
tauvec <- seq(0.20, 0.40, by=0.01)
taunum <- length(tauvec)
set.seed(332)
tempt <- proc.time()
ktemp <- Kest1k(X, xclass, alphavec, tauvec, K, 2)
proc.time()-tempt  #  23.46
ktemp$errorrate[,,3]
ktemp$device[,,3]
contour(alphavec,tauvec,ktemp$devicemat[,,3],xlab="alpha",ylab="tau",
        main="K1 Covariance",nlevels=80)
ktemp.k2 <- ktemp          # save for K2
temp <- rep(0, anum*4)     # alpha, optimal tau, deviance, error rate
dim(temp) <- c(anum, 4)
temp[,1] <- alphavec
for(i in 1:anum) {
  tauindex <- order(ktemp$devicemat[i,,3])[1];
  temp[i,2] <- tauvec[tauindex];
  temp[i,3] <- ktemp$devicemat[i, tauindex, 3];
  temp[i,4] <- ktemp$errorrate[i, tauindex, 3];
}
colnames(temp) <- c("alpha", "tau", "deviance", "error rate")
taualpha.k2 <- temp
temp
#      alpha  tau deviance error rate
# [1,]   0.0 0.30 57.15562  0.1444444
# [2,]   0.5 0.31 68.49195  0.1666667
# [3,]   1.0 0.33 77.03096  0.1666667
# [4,]   1.5 0.35 83.22773  0.1666667


## predict testing data part
## "permanental classification model", K1, K2
pre.k1 <- rep(0, tNum*tNum*anum)     # predition on testing data based on K1
dim(pre.k1) <- c(tNum*tNum, anum)
pre.k2 <- rep(0, tNum*tNum*anum)     # predition on testing data based on K2
dim(pre.k2) <- c(tNum*tNum, anum)
pre.prob.k1 <- pre.k1                # probability of true class based on K1 prediction
pre.prob.k2 <- pre.k2                # probability of true class based on K2 prediction
pre.prob1.k1 <- pre.k1               # probability of class 1 based on K1 prediction
pre.prob1.k2 <- pre.k2               # probability of class 1 based on K2 prediction

npts <- tNum                         # number of testing points, do predictions by parts
nobs <- length(xclass)               # number of training points
n <- nobs + npts
d0 <- rep(0, n*n)                    # initial distance matrix
dim(d0) <- c(n, n)
for(i in (npts+1):(n-1)) for(j in (i+1):n) {
  d0[i,j]<-d0[j,i]<-sqrt(sum((X[i-npts,]-X[j-npts,])^2));
}
dmedian <- median(d0[(npts+1):n,(npts+1):n])   # 1.544086
nclass <- max(xclass)                # number of classes


ttemp=proc.time()
for(ik in 1:tNum) {                   # loops begin to predict by parts
  Xt1 <- Xt[((ik-1)*tNum+1):(ik*tNum),];  # testing data chip in k^th loop
  tclass1 <- tclass[((ik-1)*tNum+1):(ik*tNum)]; # testing point labels in k^th loop
  d <- d0;                           # distance matrix in k^th loop
  for(i in 1:(npts-1)) for(j in (i+1):npts) {
    d[i,j]<-d[j,i]<-sqrt(sum((Xt1[i,]-Xt1[j,])^2));
  }
  for(i in 1:npts) for(j in 1:nobs) {
    d[i,j+npts]<-d[j+npts,i]<-sqrt(sum((Xt1[i,]-X[j,])^2));
  }
  d <- d/dmedian;
  # K1
  for(ia in 1:anum) {                # loops begin for different alphas
    alpha <- taualpha.k1[ia,1];
    tau <- taualpha.k1[ia,2];
    Kmat <- K1(d, tau);
    temp <- predict.self(Kmat, npts, nobs, xclass, alpha);
    pre.k1[((ik-1)*tNum+1):(ik*tNum),ia] <- temp$R3[,nclass+2];
    pre.prob.k1[((ik-1)*tNum+1):(ik*tNum),ia] <- temp$R3[cbind(seq(1:tNum),tclass1)];
    pre.prob1.k1[((ik-1)*tNum+1):(ik*tNum),ia] <- temp$R3[,1];
  }
  # K2
  for(ia in 1:anum) {                # loops begin for different alphas
    alpha <- taualpha.k2[ia,1];
    tau <- taualpha.k2[ia,2];
    Kmat <- K2(d, tau);
    temp <- predict.self(Kmat, npts, nobs, xclass, alpha);
    pre.k2[((ik-1)*tNum+1):(ik*tNum),ia] <- temp$R3[,nclass+2];
    pre.prob.k2[((ik-1)*tNum+1):(ik*tNum),ia] <- temp$R3[cbind(seq(1:tNum),tclass1)];
    pre.prob1.k2[((ik-1)*tNum+1):(ik*tNum),ia] <- temp$R3[,1];
  }
  cat("k=",ik, "\n");
}
proc.time()-ttemp # 32.77
# count error rate
error.k1 <- rep(0, anum)             # number of misclassified points based on k1
error.k2 <- rep(0, anum)             # number of misclassified points based on k2
deviance.k1 <- rep(0, anum)          # deviance based on k1
deviance.k2 <- rep(0, anum)          # deviance based on k2
for(i in 1:anum) {
  error.k1[i] <- sum(pre.k1[,i]!=tclass);
  error.k2[i] <- sum(pre.k2[,i]!=tclass);
  deviance.k1[i] <- -2*sum(log(pre.prob.k1[,i]));
  deviance.k2[i] <- -2*sum(log(pre.prob.k2[,i]));
}
temp <- cbind(error.k1, error.k2, deviance.k1, deviance.k2)
rownames(temp) <- alphavec
temp
#     error.k1 error.k2 deviance.k1 deviance.k2
# 0        305      296    1402.499    1608.184
# 0.5      325      314    2718.792    2279.753
# 1        322      314    3205.268    2694.653
# 1.5      320      312    3507.103    2984.404


## predict training data part
## "permanent classification model", K1, K2
pre.train.k1 <- rep(0, length(xclass))    # predition on training data based on K1
pre.train.k2 <- rep(0, length(xclass))    # predition on training data based on K2

npts <- nobs <- length(xclass)            # number of training points
n <- nobs + npts
d0 <- rep(0, n*n)                    # initial distance matrix
dim(d0) <- c(n, n)
for(i in (npts+1):(n-1)) for(j in (i+1):n) {
  d0[i,j]<-d0[j,i]<-sqrt(sum((X[i-npts,]-X[j-npts,])^2));
}
dmedian <- median(d0[(npts+1):n,(npts+1):n])   # 1.544086
nclass <- max(xclass)                # number of classes

Xt1 <- X
tclass1 <- xclass
d <- d0;                           # distance matrix in k^th loop
for(i in 1:(npts-1)) for(j in (i+1):npts) {
  d[i,j]<-d[j,i]<-sqrt(sum((Xt1[i,]-Xt1[j,])^2));
}
for(i in 1:npts) for(j in 1:nobs) {
  d[i,j+npts]<-d[j+npts,i]<-sqrt(sum((Xt1[i,]-X[j,])^2));
}
d <- d/dmedian;
class2 <- cbind(seq(1:npts), tclass1)

error <- matrix(0, anum, 4)       #    error.k1 error.k2 deviance.k1 deviance.k2
rownames(error) <- alphavec
colnames(error) <- c("error.k1", "error.k2", "deviance.k1", "deviance.k2")

# K1
for(ialpha in 1:anum) {        # begin of ialpha loop
alpha=alphavec[ialpha]
tau <- taualpha.k1[ialpha,2]
Kmat <- K1(d, tau)
temp <- predict.self(Kmat, npts, nobs, xclass, alpha)
pre.train.k1 <- temp$R3[,nclass+2]
error[ialpha,1] <- sum(pre.train.k1!=tclass1)
error[ialpha,3] <- -2*sum(log(temp$R3[class2]))

# K2
tau <- taualpha.k2[ialpha,2]
Kmat <- K2(d, tau)
temp <- predict.self(Kmat, npts, nobs, xclass, alpha)
pre.train.k2 <- temp$R3[,nclass+2]
error[ialpha,2] <- sum(pre.train.k2!=tclass1)
error[ialpha,4] <- -2*sum(log(temp$R3[class2]))
}                             # end of ialpha loop
error
#     error.k1 error.k2 deviance.k1 deviance.k2
# 0          0        4    9.293686    28.83411
# 0.5        0        3   46.654988    44.31106
# 1          0        3   61.700157    55.94045
# 1.5        0        3   71.482801    64.54260

## plot contour plot
xyvec <- seq(0, 3, by=3/tNum)[1:tNum] + 3/(tNum*2)
preZ3.k1 <- rep(0, tNum*tNum*anum)   # prediction probability on class 1 based on "k1"           
dim(preZ3.k1) <- c(tNum, tNum, anum)
preZ3.k2 <- preZ3.k1                 # prediction probability on class 1 based on "k1"
for(ia in 1:anum) {
  temp <- pre.prob1.k1[,ia];
  dim(temp) <- c(tNum, tNum);
  preZ3.k1[,,ia] <- t(temp);
  temp <- pre.prob1.k2[,ia];
  dim(temp) <- c(tNum, tNum);
  preZ3.k2[,,ia] <- t(temp);
}

#postscript("xc2_k1.ps",paper="special",width=8,height=7,horizontal=T)
par(mfrow=c(1,1))
contour(xyvec,xyvec,preZ3.k1[,,1],levels=c(0.5),labels=c("0"),xlab="x1",ylab="x2",
  main="Contour Plot - K1 - alpha=0, 0.5, 1", lty=1)
points(X[xclass==1,1],X[xclass==1,2],pch=1,type="p");
points(X[xclass==2,1],X[xclass==2,2],pch=4,type="p");
contour(xyvec,xyvec,preZ3.k1[,,2],levels=c(0.5),labels=c("0.5"),lty=2,add=T)
contour(xyvec,xyvec,preZ3.k1[,,3],levels=c(0.5),labels=c("1"),lty=3,add=T)
dev.off()

#postscript("xc2_k2.ps",paper="special",width=8,height=7,horizontal=T)
par(mfrow=c(1,1))
contour(xyvec,xyvec,preZ3.k2[,,1],levels=c(0.5),labels=c("0"),xlab="x1",ylab="x2",
  main="Contour Plot - K2 - alpha=0, 0.5, 1", lty=1)
points(X[xclass==1,1],X[xclass==1,2],pch=1,type="p");
points(X[xclass==2,1],X[xclass==2,2],pch=4,type="p");
contour(xyvec,xyvec,preZ3.k2[,,2],levels=c(0.5),labels=c("0.5"),lty=2,add=T)
contour(xyvec,xyvec,preZ3.k2[,,3],levels=c(0.5),labels=c("1"),lty=3,add=T)
dev.off()


# more contour lines
par(mfrow=c(1,1))
contour(xyvec,xyvec,preZ3.k1[,,1],levels=c(0.1,0.2,0.5,0.8,0.9),xlab="x1",ylab="x2",
        main="Contour Plot - K1", lty=1)
points(X[xclass==1,1],X[xclass==1,2],pch=1,type="p");
points(X[xclass==2,1],X[xclass==2,2],pch=4,type="p");


par(mfrow=c(1,1))
contour(xyvec,xyvec,preZ3.k2[,,1],levels=c(0.1,0.2,0.5,0.8,0.9),xlab="x1",ylab="x2",
        main="Contour Plot - K2", lty=1)
points(X[xclass==1,1],X[xclass==1,2],pch=1,type="p");
points(X[xclass==2,1],X[xclass==2,2],pch=4,type="p");
